'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

/* eslint prefer-rest-params: 0 */

// $FlowFixMe


var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _crypto = require('crypto');

var _crypto2 = _interopRequireDefault(_crypto);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _stream = require('stream');

var _stream2 = _interopRequireDefault(_stream);

var _url = require('url');

var _url2 = _interopRequireDefault(_url);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _async = require('async');

var _async2 = _interopRequireDefault(_async);

var _localData = require('./local-data');

var _localData2 = _interopRequireDefault(_localData);

var _streams = require('@verdaccio/streams');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var pkgFileName = 'package.json';
var fileExist = 'EEXISTS';
var noSuchFile = 'ENOENT';
var resourceNotAvailable = 'EAGAIN';

var generatePackageTemplate = function generatePackageTemplate(name) {
  return {
    // standard things
    'name': name,
    'versions': {},
    'dist-tags': {},
    'time': {},
    '_distfiles': {},
    '_attachments': {},
    '_uplinks': {}
  };
};

var DEFAULT_REVISION = `0-0000000000000000`;

/**
 * Implements Storage interface (same for storage.js, local-storage.js, up-storage.js).
 */

var Storage = function () {
  function Storage(config, logger, utils) {
    _classCallCheck(this, Storage);

    this.localData = new _localData2.default(config, logger);
    this.logger = logger.child({ sub: 'fs' });
    this.config = config;
    this.utils = utils;
  }

  _createClass(Storage, [{
    key: 'addPackage',
    value: function addPackage(name, info, callback) {
      var _this = this;

      var storage = this._getLocalStorage(name);

      if (_lodash2.default.isNil(storage)) {
        return callback(this.utils.ErrorCode.get404('this package cannot be added'));
      }

      storage.createJSON(pkgFileName, generatePackageTemplate(name), function (err) {
        if (err && err.code === fileExist) {
          return callback(_this.utils.ErrorCode.get409());
        }

        var latest = _this.utils.getLatestVersion(info);

        if (_lodash2.default.isNil(latest) === false && info.versions[latest]) {
          return callback(null, info.versions[latest]);
        }
        return callback();
      });
    }

    /**
     * Remove package.
     * @param {*} name
     * @param {*} callback
     * @return {Function}
     */

  }, {
    key: 'removePackage',
    value: function removePackage(name, callback) {
      var _this2 = this;

      var storage = this._getLocalStorage(name);
      if (_lodash2.default.isNil(storage)) {
        return callback(this.utils.ErrorCode.get404());
      }

      storage.readJSON(pkgFileName, function (err, data) {
        if (_lodash2.default.isNil(err) === false) {
          if (err.code === noSuchFile) {
            return callback(_this2.utils.ErrorCode.get404());
          } else {
            return callback(err);
          }
        }
        _this2._normalizePackage(data);

        var removeFailed = _this2.localData.remove(name);

        if (removeFailed) {
          // This will happen when database is locked
          return callback(_this2.utils.ErrorCode.get422(removeFailed.message));
        }

        storage.deleteJSON(pkgFileName, function (err) {
          if (err) {
            return callback(err);
          }
          var attachments = Object.keys(data._attachments);

          var unlinkNext = function unlinkNext(cb) {
            if (_lodash2.default.isEmpty(attachments)) {
              return cb();
            }

            var attachment = attachments.shift();
            storage.deleteJSON(attachment, function () {
              unlinkNext(cb);
            });
          };

          unlinkNext(function () {
            // try to unlink the directory, but ignore errors because it can fail
            storage.removePackage('.', function (err) {
              callback(err);
            });
          });
        });
      });
    }

    /**
     * Synchronize remote package info with the local one
     * @param {*} name
     * @param {*} packageInfo
     * @param {*} callback
     */

  }, {
    key: 'updateVersions',
    value: function updateVersions(name, packageInfo, callback) {
      var _this3 = this;

      this._readCreatePackage(name, function (err, packageLocalJson) {
        if (err) {
          return callback(err);
        }

        var change = false;
        for (var versionId in packageInfo.versions) {
          if (_lodash2.default.isNil(packageLocalJson.versions[versionId])) {
            var version = packageInfo.versions[versionId];

            // we don't keep readmes for package versions,
            // only one readme per package
            delete version.readme;

            change = true;
            packageLocalJson.versions[versionId] = version;

            if (version.dist && version.dist.tarball) {
              var urlObject = _url2.default.parse(version.dist.tarball);
              var filename = urlObject.pathname.replace(/^.*\//, '');

              // we do NOT overwrite any existing records
              if (_lodash2.default.isNil(packageLocalJson._distfiles[filename])) {
                var hash = packageLocalJson._distfiles[filename] = {
                  url: version.dist.tarball,
                  sha: version.dist.shasum
                };
                // $FlowFixMe
                var upLink = version[Symbol.for('__verdaccio_uplink')];

                if (_lodash2.default.isNil(upLink) === false) {
                  _this3._updateUplinkToRemoteProtocol(hash, upLink);
                }
              }
            }
          }
        }
        for (var tag in packageInfo['dist-tags']) {
          if (!packageLocalJson['dist-tags'][tag] || packageLocalJson['dist-tags'][tag] !== packageInfo['dist-tags'][tag]) {
            change = true;
            packageLocalJson['dist-tags'][tag] = packageInfo['dist-tags'][tag];
          }
        }
        for (var up in packageInfo._uplinks) {
          if (Object.prototype.hasOwnProperty.call(packageInfo._uplinks, up)) {
            var need_change = !_this3.utils.is_object(packageLocalJson._uplinks[up]) || packageInfo._uplinks[up].etag !== packageLocalJson._uplinks[up].etag || packageInfo._uplinks[up].fetched !== packageLocalJson._uplinks[up].fetched;

            if (need_change) {
              change = true;
              packageLocalJson._uplinks[up] = packageInfo._uplinks[up];
            }
          }
        }

        if (packageInfo.readme !== packageLocalJson.readme) {
          packageLocalJson.readme = packageInfo.readme;
          change = true;
        }

        if ('time' in packageInfo) {
          packageLocalJson.time = packageInfo.time;
          change = true;
        }

        if (change) {
          _this3.logger.debug('updating package info');
          _this3._writePackage(name, packageLocalJson, function (err) {
            callback(err, packageLocalJson);
          });
        } else {
          callback(null, packageLocalJson);
        }
      });
    }

    /**
     * Ensure the dist file remains as the same protocol
     * @param {Object} hash metadata
     * @param {String} upLink registry key
     * @private
     */

  }, {
    key: '_updateUplinkToRemoteProtocol',
    value: function _updateUplinkToRemoteProtocol(hash, upLinkKey) {
      // if we got this information from a known registry,
      // use the same protocol for the tarball
      //
      var tarballUrl = _url2.default.parse(hash.url);
      var uplinkUrl = _url2.default.parse(this.config.uplinks[upLinkKey].url);

      if (uplinkUrl.host === tarballUrl.host) {
        tarballUrl.protocol = uplinkUrl.protocol;
        hash.registry = upLinkKey;
        hash.url = _url2.default.format(tarballUrl);
      }
    }

    /**
     * Add a new version to a previous local package.
     * @param {*} name
     * @param {*} version
     * @param {*} metadata
     * @param {*} tag
     * @param {*} callback
     */

  }, {
    key: 'addVersion',
    value: function addVersion(name, version, metadata, tag, callback) {
      var _this4 = this;

      this._updatePackage(name, function (data, cb) {
        // keep only one readme per package
        data.readme = metadata.readme;

        // TODO: lodash remove
        delete metadata.readme;

        if (data.versions[version] != null) {
          return cb(_this4.utils.ErrorCode.get409());
        }

        // if uploaded tarball has a different shasum, it's very likely that we have some kind of error
        if (_this4.utils.is_object(metadata.dist) && _lodash2.default.isString(metadata.dist.tarball)) {
          var tarball = metadata.dist.tarball.replace(/.*\//, '');

          if (_this4.utils.is_object(data._attachments[tarball])) {

            if (_lodash2.default.isNil(data._attachments[tarball].shasum) === false && _lodash2.default.isNil(metadata.dist.shasum) === false) {
              if (data._attachments[tarball].shasum != metadata.dist.shasum) {
                var errorMessage = `shasum error, ${data._attachments[tarball].shasum} != ${metadata.dist.shasum}`;
                return cb(_this4.utils.ErrorCode.get400(errorMessage));
              }
            }

            var currentDate = new Date().toISOString();
            data.time['modified'] = currentDate;

            if ('created' in data.time === false) {
              data.time.created = currentDate;
            }

            data.time[version] = currentDate;
            data._attachments[tarball].version = version;
          }
        }

        data.versions[version] = metadata;
        _this4.utils.tag_version(data, version, tag);

        var addFailed = _this4.localData.add(name);
        if (addFailed) {
          return cb(_this4.utils.ErrorCode.get422(addFailed.message));
        }

        cb();
      }, callback);
    }

    /**
     * Merge a new list of tags for a local packages with the existing one.
     * @param {*} name
     * @param {*} tags
     * @param {*} callback
     */

  }, {
    key: 'mergeTags',
    value: function mergeTags(name, tags, callback) {
      var _this5 = this;

      this._updatePackage(name, function (data, cb) {
        for (var t in tags) {
          if (_lodash2.default.isNull(tags[t])) {
            delete data['dist-tags'][t];
            continue;
          }

          if (_lodash2.default.isNil(data.versions[tags[t]])) {
            return cb(_this5._getVersionNotFound());
          }
          var key = tags[t];
          _this5.utils.tag_version(data, key, t);
        }
        cb();
      }, callback);
    }

    /**
     * Return version not found
     * @return {String}
     * @private
     */

  }, {
    key: '_getVersionNotFound',
    value: function _getVersionNotFound() {
      return this.utils.ErrorCode.get404('this version doesn\'t exist');
    }
    /**
     * Return file no available
     * @return {String}
     * @private
     */

  }, {
    key: '_getFileNotAvailable',
    value: function _getFileNotAvailable() {
      return this.utils.ErrorCode.get404('no such file available');
    }

    /**
     * Update the package metadata, tags and attachments (tarballs).
     * Note: Currently supports unpublishing only.
     * @param {*} name
     * @param {*} metadata
     * @param {*} revision
     * @param {*} callback
     * @return {Function}
     */

  }, {
    key: 'changePackage',
    value: function changePackage(name, metadata, revision, callback) {
      var _this6 = this;

      if (!this.utils.is_object(metadata.versions) || !this.utils.is_object(metadata['dist-tags'])) {
        return callback(this.utils.ErrorCode.get422());
      }

      this._updatePackage(name, function (data, cb) {
        for (var ver in data.versions) {
          if (_lodash2.default.isNil(metadata.versions[ver])) {
            _this6.logger.info({ name: name, version: ver }, 'unpublishing @{name}@@{version}');
            delete data.versions[ver];
            for (var file in data._attachments) {
              if (data._attachments[file].version === ver) {
                delete data._attachments[file].version;
              }
            }
          }
        }
        data['dist-tags'] = metadata['dist-tags'];
        cb();
      }, function (err) {
        if (err) {
          return callback(err);
        }
        callback();
      });
    }

    /**
     * Remove a tarball.
     * @param {*} name
     * @param {*} filename
     * @param {*} revision
     * @param {*} callback
     */

  }, {
    key: 'removeTarball',
    value: function removeTarball(name, filename, revision, callback) {
      var _this7 = this;

      (0, _assert2.default)(this.utils.validate_name(filename));

      this._updatePackage(name, function (data, cb) {
        if (data._attachments[filename]) {
          delete data._attachments[filename];
          cb();
        } else {
          cb(_this7._getFileNotAvailable());
        }
      }, function (err) {
        if (err) {
          return callback(err);
        }
        var storage = _this7._getLocalStorage(name);

        if (storage) {
          storage.deleteJSON(filename, callback);
        }
      });
    }

    /**
     * Add a tarball.
     * @param {String} name
     * @param {String} filename
     * @return {Stream}
     */

  }, {
    key: 'addTarball',
    value: function addTarball(name, filename) {
      var _this8 = this;

      (0, _assert2.default)(this.utils.validate_name(filename));

      var length = 0;
      var shaOneHash = _crypto2.default.createHash('sha1');
      var uploadStream = new _streams.UploadTarball();
      var _transform = uploadStream._transform;
      var storage = this._getLocalStorage(name);
      uploadStream.abort = function () {};
      uploadStream.done = function () {};

      uploadStream._transform = function (data) {
        shaOneHash.update(data);
        // measure the length for validation reasons
        length += data.length;
        _transform.apply(uploadStream, arguments);
      };

      if (name === pkgFileName || name === '__proto__') {
        process.nextTick(function () {
          uploadStream.emit('error', _this8.utils.ErrorCode.get403());
        });
        return uploadStream;
      }

      if (!storage) {
        process.nextTick(function () {
          uploadStream.emit('error', 'can\'t upload this package');
        });
        return uploadStream;
      }

      var writeStream = storage.createWriteStream(filename);

      writeStream.on('error', function (err) {
        if (err.code === fileExist) {
          uploadStream.emit('error', _this8.utils.ErrorCode.get409());
        } else if (err.code === noSuchFile) {
          // check if package exists to throw an appropriate message
          _this8.getPackageMetadata(name, function (_err, res) {
            if (_err) {
              uploadStream.emit('error', _err);
            } else {
              uploadStream.emit('error', err);
            }
          });
        } else {
          uploadStream.emit('error', err);
        }
      });

      writeStream.on('open', function () {
        // re-emitting open because it's handled in storage.js
        uploadStream.emit('open');
      });

      writeStream.on('success', function () {
        _this8._updatePackage(name, function updater(data, cb) {
          data._attachments[filename] = {
            shasum: shaOneHash.digest('hex')
          };
          cb();
        }, function (err) {
          if (err) {
            uploadStream.emit('error', err);
          } else {
            uploadStream.emit('success');
          }
        });
      });

      uploadStream.abort = function () {
        writeStream.abort();
      };

      uploadStream.done = function () {
        if (!length) {
          uploadStream.emit('error', this.utils.ErrorCode.get422('refusing to accept zero-length file'));
          writeStream.abort();
        } else {
          writeStream.done();
        }
      };

      uploadStream.pipe(writeStream);

      return uploadStream;
    }

    /**
     * Get a tarball.
     * @param {*} name
     * @param {*} filename
     * @return {ReadTarball}
     */

  }, {
    key: 'getTarball',
    value: function getTarball(name, filename) {
      (0, _assert2.default)(this.utils.validate_name(filename));

      var storage = this._getLocalStorage(name);

      if (_lodash2.default.isNil(storage)) {
        return this._createFailureStreamResponse();
      }

      return this._streamSuccessReadTarBall(storage, filename);
    }

    /**
     * Return a stream that emits a read failure.
     * @private
     * @return {ReadTarball}
     */

  }, {
    key: '_createFailureStreamResponse',
    value: function _createFailureStreamResponse() {
      var _this9 = this;

      var stream = new _streams.ReadTarball();

      process.nextTick(function () {
        stream.emit('error', _this9._getFileNotAvailable());
      });
      return stream;
    }

    /**
     * Return a stream that emits the tarball data
     * @param {Object} storage
     * @param {String} filename
     * @private
     * @return {ReadTarball}
     */

  }, {
    key: '_streamSuccessReadTarBall',
    value: function _streamSuccessReadTarBall(storage, filename) {
      var stream = new _streams.ReadTarball();
      var readTarballStream = storage.createReadStream(filename);
      var e404 = this.utils.ErrorCode.get404;

      stream.abort = function () {
        if (_lodash2.default.isNil(readTarballStream) === false) {
          readTarballStream.abort();
        }
      };

      readTarballStream.on('error', function (err) {
        if (err && err.code === noSuchFile) {
          stream.emit('error', e404('no such file available'));
        } else {
          stream.emit('error', err);
        }
      });

      readTarballStream.on('content-length', function (v) {
        stream.emit('content-length', v);
      });

      readTarballStream.on('open', function () {
        // re-emitting open because it's handled in storage.js
        stream.emit('open');
        readTarballStream.pipe(stream);
      });

      return stream;
    }

    /**
     * Retrieve a package by name.
     * @param {*} name
     * @param {*} options
     * @param {*} callback
     * @return {Function}
     */

  }, {
    key: 'getPackageMetadata',
    value: function getPackageMetadata(name) {
      var callback = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function () {};


      var storage = this._getLocalStorage(name);
      if (_lodash2.default.isNil(storage)) {
        return callback(this.utils.ErrorCode.get404());
      }

      this._readJSON(storage, callback);
    }

    /**
     * Search a local package.
     * @param {*} startKey
     * @param {*} options
     * @return {Function}
     */

  }, {
    key: 'search',
    value: function search(startKey, options) {
      var _this10 = this;

      var stream = new _stream2.default.PassThrough({ objectMode: true });

      this._eachPackage(function (item, cb) {
        _fs2.default.stat(item.path, function (err, stats) {
          if (_lodash2.default.isNil(err) === false) {
            return cb(err);
          }

          if (stats.mtime.getTime() > parseInt(startKey, 10)) {
            _this10.getPackageMetadata(item.name, function (err, data) {
              if (err) {
                return cb(err);
              }
              var listVersions = Object.keys(data.versions);
              var versions = _this10.utils.semver_sort(listVersions);
              var latest = data['dist-tags'] && data['dist-tags'].latest ? data['dist-tags'].latest : versions.pop();

              if (data.versions[latest]) {
                var version = data.versions[latest];
                var pkg = {
                  'name': version.name,
                  'description': version.description,
                  'dist-tags': { latest },
                  'maintainers': version.maintainers || [version.author].filter(Boolean),
                  'author': version.author,
                  'repository': version.repository,
                  'readmeFilename': version.readmeFilename || '',
                  'homepage': version.homepage,
                  'keywords': version.keywords,
                  'bugs': version.bugs,
                  'license': version.license,
                  'time': {
                    modified: item.time ? new Date(item.time).toISOString() : stats.mtime
                  },
                  'versions': { [latest]: 'latest' }
                };

                stream.push(pkg);
              }

              cb();
            });
          } else {
            cb();
          }
        });
      }, function onEnd(err) {
        if (err) {
          return stream.emit('error', err);
        }
        stream.end();
      });

      return stream;
    }

    /**
     * Retrieve a wrapper that provide access to the package location.
     * @param {Object} packageInfo package name.
     * @return {Object}
     */

  }, {
    key: '_getLocalStorage',
    value: function _getLocalStorage(packageInfo) {
      var path = this.__getLocalStoragePath(this.config.getMatchedPackagesSpec(packageInfo).storage);

      if (_lodash2.default.isNil(path) || path === false) {
        this.logger.debug({ name: packageInfo }, 'this package has no storage defined: @{name}');
        return null;
      }

      return this.localData.getPackageStorage(packageInfo, path);
    }

    /**
     * Read a json file from storage.
     * @param {Object} storage
     * @param {Function} callback
     */

  }, {
    key: '_readJSON',
    value: function _readJSON(storage, callback) {
      var _this11 = this;

      storage.readJSON(pkgFileName, function (err, result) {
        if (err) {
          if (err.code === noSuchFile) {
            return callback(_this11.utils.ErrorCode.get404());
          } else {
            return callback(_this11._internalError(err, pkgFileName, 'error reading'));
          }
        }
        _this11._normalizePackage(result);
        callback(err, result);
      });
    }

    /**
     * Verify the right local storage location.
     * @param {String} path
     * @return {String}
     * @private
     */

  }, {
    key: '__getLocalStoragePath',
    value: function __getLocalStoragePath(path) {
      if (_lodash2.default.isNil(path)) {
        path = this.config.storage;
      }
      return path;
    }

    /**
     * Walks through each package and calls `on_package` on them.
     * @param {*} onPackage
     * @param {*} on_end
     */

  }, {
    key: '_eachPackage',
    value: function _eachPackage(onPackage, on_end) {
      var _this12 = this;

      var storages = {};
      var utils = this.utils;

      storages[this.config.storage] = true;
      if (this.config.packages) {
        Object.keys(this.config.packages || {}).map(function (pkg) {
          if (_this12.config.packages[pkg].storage) {
            storages[_this12.config.packages[pkg].storage] = true;
          }
        });
      }
      var base = _path2.default.dirname(this.config.self_path);

      _async2.default.eachSeries(Object.keys(storages), function (storage, cb) {
        _fs2.default.readdir(_path2.default.resolve(base, storage), function (err, files) {
          if (err) {
            return cb(err);
          }

          _async2.default.eachSeries(files, function (file, cb) {
            if (file.match(/^@/)) {
              // scoped
              _fs2.default.readdir(_path2.default.resolve(base, storage, file), function (err, files) {
                if (err) {
                  return cb(err);
                }

                _async2.default.eachSeries(files, function (file2, cb) {
                  if (utils.validate_name(file2)) {
                    onPackage({
                      name: `${file}/${file2}`,
                      path: _path2.default.resolve(base, storage, file, file2)
                    }, cb);
                  } else {
                    cb();
                  }
                }, cb);
              });
            } else if (utils.validate_name(file)) {
              onPackage({
                name: file,
                path: _path2.default.resolve(base, storage, file)
              }, cb);
            } else {
              cb();
            }
          }, cb);
        });
      }, on_end);
    }

    /**
     * Normalise package properties, tags, revision id.
     * @param {Object} pkg package reference.
     */

  }, {
    key: '_normalizePackage',
    value: function _normalizePackage(pkg) {
      var _this13 = this;

      var pkgProperties = ['versions', 'dist-tags', '_distfiles', '_attachments', '_uplinks', 'time'];

      pkgProperties.forEach(function (key) {
        if (!_this13.utils.is_object(pkg[key])) {
          pkg[key] = {};
        }
      });

      if (_lodash2.default.isString(pkg._rev) === false) {
        pkg._rev = DEFAULT_REVISION;
      }
      // normalize dist-tags
      this.utils.normalize_dist_tags(pkg);
    }

    /**
     * Retrieve either a previous created local package or a boilerplate.
     * @param {*} name
     * @param {*} callback
     * @return {Function}
     */

  }, {
    key: '_readCreatePackage',
    value: function _readCreatePackage(name, callback) {
      var _this14 = this;

      var storage = this._getLocalStorage(name);
      if (!storage) {
        var data = generatePackageTemplate(name);
        this._normalizePackage(data);
        return callback(null, data);
      }

      storage.readJSON(pkgFileName, function (err, data) {
        // TODO: race condition
        if (err) {
          if (err.code === noSuchFile) {
            // if package doesn't exist, we create it here
            data = generatePackageTemplate(name);
          } else {
            return callback(_this14._internalError(err, pkgFileName, 'error reading'));
          }
        }
        _this14._normalizePackage(data);
        callback(null, data);
      });
    }

    /**
     * Handle internal error
     * @param {*} err
     * @param {*} file
     * @param {*} message
     * @return {Object} Error instance
     */

  }, {
    key: '_internalError',
    value: function _internalError(err, file, message) {
      this.logger.error({ err: err, file: file }, message + ' @{file}: @{!err.message}');
      return this.utils.ErrorCode.get500();
    }

    /**
     * This function allows to update the package thread-safely
     Algorithm:
     1. lock package.json for writing
     2. read package.json
     3. updateFn(pkg, cb), and wait for cb
     4. write package.json.tmp
     5. move package.json.tmp package.json
     6. callback(err?)
     * @param {*} name package name
     * @param {*} updateFn function(package, cb) - update function
     * @param {*} _callback callback that gets invoked after it's all updated
     * @return {Function}
     */

  }, {
    key: '_updatePackage',
    value: function _updatePackage(name, updateFn, _callback) {
      var _this15 = this;

      var storage = this._getLocalStorage(name);
      if (!storage) {
        return _callback(this.utils.ErrorCode.get404());
      }
      storage.lockAndReadJSON(pkgFileName, function (err, json) {
        var locked = false;

        // callback that cleans up lock first
        var callback = function callback(err) {
          var _args = arguments;
          if (locked) {
            storage.unlock_file(pkgFileName, function () {
              // ignore any error from the unlock
              _callback.apply(err, _args);
            });
          } else {
            _callback.apply(undefined, _toConsumableArray(_args));
          }
        };

        if (!err) {
          locked = true;
        }

        if (err) {
          if (err.code === resourceNotAvailable) {
            return callback(_this15.utils.ErrorCode.get503());
          } else if (err.code === noSuchFile) {
            return callback(_this15.utils.ErrorCode.get404());
          } else {
            return callback(err);
          }
        }

        _this15._normalizePackage(json);
        updateFn(json, function (err) {
          if (err) {
            return callback(err);
          }
          _this15._writePackage(name, json, callback);
        });
      });
    }

    /**
     * Update the revision (_rev) string for a package.
     * @param {*} name
     * @param {*} json
     * @param {*} callback
     * @return {Function}
     */

  }, {
    key: '_writePackage',
    value: function _writePackage(name, json, callback) {
      // calculate revision a la couchdb
      if (typeof json._rev !== 'string') {
        json._rev = DEFAULT_REVISION;
      }

      json._rev = this._generateRevision(json._rev);

      var storage = this._getLocalStorage(name);
      if (_lodash2.default.isNil(storage)) {
        return callback();
      }
      storage.writeJSON(pkgFileName, json, callback);
    }
  }, {
    key: '_generateRevision',
    value: function _generateRevision(rev) {
      var _rev = rev.split('-');

      return (+_rev[0] || 0) + 1 + '-' + _crypto2.default.pseudoRandomBytes(8).toString('hex');
    }
  }]);

  return Storage;
}();

module.exports = Storage;